import torch
import torch.autograd as autograd
import torch.nn as nn

import numpy as np
from pyDOE import lhs
import scipy.io

import matplotlib.pyplot as plt
#%%
'''初期設定'''
torch.set_default_dtype(torch.float64)

torch.manual_seed(123)
np.random.seed(123)

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print(device)

if device == 'cuda': 
    print(torch.cuda.get_device_name())

c=2 #wave velocity
#%%
'''理論解の読み込み'''
data = scipy.io.loadmat('1Dcontinuum_vib_1stmode.mat')
x = data['x'] #dim: 1 x Nx
t = data['t'] #dim: 1 x Nt
Nx = x.shape[1]
Nt = t.shape[1]
dx = x[0, 1]
dt = t[0, 1]

uexact = data['uexact'] #dim: Nx x Nt
udot = data['udot'] #dim: Nx x Nt
udash = data['udash'] #dim: Nx x Nt

X, T = np.meshgrid(x, t)
xt_test = np.hstack((X.flatten()[:,None], T.flatten()[:,None]))

# Domain bounds
lb = xt_test[0]  # [0. 0.]
ub = xt_test[-1] # [L, T]

u_true = uexact.flatten('F')[:,None]
#%%
'''評価点の設定'''
def trainingdata(N_uIni, N_uDir, N_uNeu, N_f):

    #Boundary Condition x = 0 and 0 <= t <= T: Dirichlet conditions (x=...)
    dirichlet_xt = np.hstack((X[:,0][:,None], T[:,0][:,None]))
    dirichlet_u = uexact[0,:][:,None]

    #Boundary Condition x = L and 0 <= t <= T: Neumann conditions (x'=...)
    neumann_xt = np.hstack((X[:,-1][:,None], T[:,0][:,None])) #L3
    neumann_udash = udash[-1,:][:,None]

    #Initial Condition 0 <= x <= L and t = 0
    initial_xt = np.hstack((X[0,:][:,None], T[0,:][:,None]))
    initial_u = uexact[:,0][:,None]
    initial_udot = udot[:,0][:,None]

    #choose random N_u points for training boundary conditions
    idx_ini = np.random.choice(initial_xt.shape[0], N_uIni, replace=False) 
    idx_dir = np.random.choice(dirichlet_xt.shape[0], N_uDir, replace=False)
    idx_neu = np.random.choice(neumann_xt.shape[0], N_uNeu, replace=False)

    xt_ini_train = initial_xt[idx_ini, :]
    u_ini_train = initial_u[idx_ini,:]
    udot_ini_train = initial_udot[idx_ini,:]

    xt_dir_train = dirichlet_xt[idx_dir,:]
    u_dir_train = dirichlet_u[idx_dir,:]
    
    xt_neu_train = neumann_xt[idx_neu,:]
    u_neu_train = neumann_udash[idx_neu,:]

    #Collcation point generation by Latin Hypercube Sampling
    xt_f_train = lb + (ub-lb)*lhs(2,N_f) 

    return xt_ini_train,u_ini_train,udot_ini_train,xt_dir_train,u_dir_train,xt_neu_train,u_neu_train,xt_f_train

N_uIni = 1000
N_uDir = 100
N_uNeu = 100
N_f = 10000
xt_ini_train_np, u_ini_train_np, udot_ini_train_np, xt_dir_train_np, u_dir_train_np, xt_neu_train_np, u_neu_train_np, xt_f_train_np = trainingdata(N_uIni, N_uDir, N_uNeu, N_f)

#Convert to tensor and send to GPU
xt_ini_train = torch.from_numpy(xt_ini_train_np).double().to(device)
u_ini_train = torch.from_numpy(u_ini_train_np).double().to(device)
udot_ini_train = torch.from_numpy(udot_ini_train_np).double().to(device)
xt_dir_train = torch.from_numpy(xt_dir_train_np).double().to(device)
u_dir_train = torch.from_numpy(u_dir_train_np).double().to(device)
xt_neu_train = torch.from_numpy(xt_neu_train_np).double().to(device)
u_neu_train = torch.from_numpy(u_neu_train_np).double().to(device)
xt_f_train =  torch.from_numpy(xt_f_train_np).double().to(device)

xt_test_tensor = torch.from_numpy(xt_test).double().to(device)
u = torch.from_numpy(u_true).double().to(device)

f_hat = torch.zeros(xt_f_train.shape[0],1).to(device)
#%%
'''モデル設計'''
class Sequentialmodel(nn.Module):
    
    def __init__(self,layers):
        super().__init__()
              
        'activation function'
        self.activation = nn.Tanh()

        'loss function'
        self.loss_function = nn.MSELoss(reduction ='mean')
    
        'Initialise neural network as a list using nn.Modulelist'  
        self.linears = nn.ModuleList([nn.Linear(layers[i], layers[i+1]) for i in range(len(layers)-1)])
        
        self.iter = 0
        
        self.loss_hist = []

        for i in range(len(layers)-1):
            nn.init.xavier_normal_(self.linears[i].weight.data, gain=1.0)
            nn.init.zeros_(self.linears[i].bias.data)
            
    #foward computation
    def forward(self,x):
        
        if torch.is_tensor(x) != True:         
            x = torch.from_numpy(x)           
        
        u_b = torch.from_numpy(ub).double().to(device)
        l_b = torch.from_numpy(lb).double().to(device)
                      
        #scaling
        x = (x - l_b)/(u_b - l_b)
        
        #convert to double
        a = x.double()

        for i in range(len(layers)-2):         
            z = self.linears[i](a)
            a = self.activation(z)

        a = self.linears[-1](a)
        
        return a

    #loss function induced from initial condition
    def loss_IC(self,xt,y,y_t):

        z = xt.clone()
        z.requires_grad = True
        
        u = self.forward(z)
        u_x_t = autograd.grad(u,z,torch.ones([xt.shape[0], 1]).to(device), retain_graph=True, create_graph=True)[0]
        u_t = u_x_t[:,[1]]  

        loss_u = self.loss_function(u, y) + self.loss_function(u_t, y_t)
                
        return loss_u

    #loss function induced from Diriclet boundary condition
    def loss_BC_Diriclet(self,xt,y):
                
        loss_u = self.loss_function(self.forward(xt), y)
                
        return loss_u

    #loss function induced from Neumann boundary condition
    def loss_BC_Neumann(self,xt,y_x):
                
        z = xt.clone()       
        z.requires_grad = True
        u = self.forward(z)
        u_x_t = autograd.grad(u,z,torch.ones([xt.shape[0], 1]).to(device), retain_graph=True, create_graph=True)[0]
        u_x = u_x_t[:,[0]]      

        loss_u = self.loss_function(u_x, y_x)
                
        return loss_u

    #loss function induced from governing equation
    def loss_PDE(self, xt):
                       
        z = xt.clone()
        z.requires_grad = True
        
        u = self.forward(z)
        u_x_t = autograd.grad(u,z,torch.ones([xt.shape[0], 1]).to(device), retain_graph=True, create_graph=True)[0]
        u_x = u_x_t[:, [0]]
        u_t = u_x_t[:, [1]]  
        u_xx_xt = autograd.grad(u_x,z,torch.ones([xt.shape[0], 1]).to(device), create_graph=True)[0]
        u_tx_tt = autograd.grad(u_t,z,torch.ones([xt.shape[0], 1]).to(device), create_graph=True)[0]
        u_xx = u_xx_xt[:,[0]]
        u_tt = u_tx_tt[:,[1]]
        
        f = u_tt - c**2 * u_xx
        loss_f = self.loss_function(f,f_hat)
                
        return loss_f

    #summation of loss functions
    def loss(self,xt_ini_train,u_ini_train,udot_ini_train,xt_dir_train,u_dir_train,xt_neu_train,u_neu_train,xt_f_train):

        loss_ini = self.loss_IC(xt_ini_train, u_ini_train, udot_ini_train)        
        loss_BC_Diriclet = self.loss_BC_Diriclet(xt_dir_train, u_dir_train)
        loss_BC_Neumann = self.loss_BC_Neumann(xt_neu_train, u_neu_train)
        
        loss_f = self.loss_PDE(xt_f_train)
        
        loss_val = loss_ini + loss_BC_Diriclet + loss_BC_Neumann + 10*loss_f
        
        return loss_val
                                       
    def closure(self):
        
        optimizer.zero_grad()
        loss = self.loss(xt_ini_train,u_ini_train,udot_ini_train,xt_dir_train,u_dir_train,xt_neu_train,u_neu_train,xt_f_train)
        self.loss_hist.append(loss.item())
        
        loss.backward()
                
        self.iter += 1
        
        if self.iter % 100 == 0:
            _ = PINN.test()
            print('step=',self.iter, ', loss=', loss.item())

        return loss        
    
    def test(self):
                
        u_pred = self.forward(xt_test_tensor)
        u_pred = u_pred.cpu().detach().numpy()
        u_pred = np.reshape(u_pred,(Nx,Nt),order='F')
                
        return u_pred

    def test_grad(self):

        z = xt_test_tensor.clone()
        z.requires_grad = True
        
        u = self.forward(z)
        u_x_t = autograd.grad(u,z,torch.ones([xt_test_tensor.shape[0], 1]).to(device), retain_graph=True, create_graph=True)[0]
        u_x = u_x_t[:,[0]]        

        u_pred = u_x.cpu().detach().numpy()
        u_pred = np.reshape(u_pred,(Nx,Nt),order='F')
                
        return u_pred

    def return_loss(self):
        z = xt_test_tensor.clone()
        z.requires_grad = True
        
        u = self.forward(z)
        u_x_t = autograd.grad(u,z,torch.ones([xt_test_tensor.shape[0], 1]).to(device), retain_graph=True, create_graph=True)[0]
        u_xx_tt = autograd.grad(u_x_t,z,torch.ones(xt_test_tensor.shape).to(device), create_graph=True)[0]
        u_xx = u_xx_tt[:,[0]]
        u_tt = u_xx_tt[:,[1]]
        
        f = u_tt - c**2 * u_xx
        f = f.cpu().detach().numpy()
        f = np.reshape(f, (Nx, Nt), order='F')
    
        return f

layers = np.array([2,20,20,20,20,20,20,20,20,1])

PINN = Sequentialmodel(layers)
PINN.to(device)

print(PINN)
#%%
'''最適化'''
params = list(PINN.parameters())

optimizer = torch.optim.LBFGS(PINN.parameters(), lr=1.0, 
                              max_iter = 10000, 
                              max_eval = 100000, 
                              tolerance_grad = 1e-10, 
                              tolerance_change = 1e-10, 
                              history_size = 100, 
                              line_search_fn = 'strong_wolfe')

optimizer.step(PINN.closure)
hist_lbfgs = PINN.loss_hist
#%%
'''解の可視化'''
upred = PINN.test()
ugrad_x = PINN.test_grad()

x_eval = np.ravel(X.T).reshape(-1, 1)
t_eval = np.ravel(T.T).reshape(-1, 1)

xmax = np.max(x_eval)
tmax = np.max(t_eval)

plt.figure()
plt.xlim(0, tmax)
plt.ylim(0, xmax)
plt.xlabel('t')
plt.ylabel('x')
plt.scatter(t_eval, x_eval, c=uexact, vmin=-1, vmax=1, cmap='rainbow')
plt.colorbar()
plt.title('exact solution')
plt.show()


plt.figure()
plt.xlim(0, tmax)
plt.ylim(0, xmax)
plt.xlabel('t')
plt.ylabel('x')
plt.scatter(t_eval, x_eval, c=upred, vmin=-1, vmax=1, cmap='rainbow')
plt.colorbar()
plt.title('PINNs solution')
plt.show()

plotx = [0.00, 0.25, 0.5, 1.0]
plt.figure()
plt.xlim(0, tmax)
plt.ylim(-1.1, 1.1)
plt.xlabel('t')
plt.ylabel('u')
for x in plotx:
    idx = int(x/dx)
    
    if x == plotx[0]:
        plt.plot(np.arange(Nt)*dt, upred[idx, :], color='blue', label='PINNs sol.')
        plt.plot(np.arange(Nt)*dt, uexact[idx, :], linestyle='dashed', color='red', label='exact sol.') 
    else:
        plt.plot(np.arange(Nt)*dt, upred[idx, :], color='blue')
        plt.plot(np.arange(Nt)*dt, uexact[idx, :], linestyle='dashed', color='red') 
    plt.text(t[0, -1]+dt, uexact[idx, -1], 'x={}'.format(x))
    plt.legend(loc='best')
#%%
"""lossの可視化"""
loss_f = PINN.return_loss()
plt.figure()
plt.xlim(0, tmax)
plt.ylim(0, xmax)
plt.xlabel('t')
plt.ylabel('x')
plt.scatter(t_eval, x_eval, c=loss_f, vmin=-0.01, vmax=0.01, cmap='rainbow')
plt.colorbar()
plt.title('loss_pde [c={}]'.format(c))
plt.show()

loss_f = PINN.return_loss()
plt.figure()
plt.xlim(0, 1)
plt.ylim(0, xmax)
plt.xlabel('t')
plt.ylabel('x')
plt.scatter(t_eval, x_eval, c=loss_f, vmin=-10, vmax=10, cmap='rainbow')
plt.colorbar()
plt.title('loss_pde [c={}]'.format(c))
plt.show()

plt.figure()
plt.imshow(loss_f)
plt.colorbar()

plt.figure()
plt.figure()
plt.xlim(0, tmax)
plt.ylim(0, xmax)
plt.xlabel('t')
plt.ylabel('x')
plt.scatter(t_eval, x_eval, c=uexact-upred, vmin=-0.01, vmax=0.01, cmap='rainbow')
plt.colorbar()
plt.title('regression loss [c={}]'.format(c))
plt.show()


plt.figure()
plt.figure()
plt.xlim(0, tmax)
plt.ylim(0, xmax)
plt.xlabel('t')
plt.ylabel('x')
plt.scatter(t_eval, x_eval, c=udash-ugrad_x, vmin=-0.01, vmax=0.01, cmap='rainbow')
plt.colorbar()
plt.title('regression loss (u_x) [c={}]'.format(c))
plt.show()